#include <stdio.h>
#include "parti.h"

void 
new_multi_execute_gather(pattern, size, narray, buf_list, data)
struct com_pattern *pattern;		/* communication pattern */
int  size;				/* size of array elements */
int  narray;				/* number of arrays */
char **buf_list,**data;			/* source and destination arrays */
{
  static int call_count = 0;		/* counter for message ids */
  int  node;				/* processor id of incoming message */
  int  allocated = FALSE;		/* new buffer allocated */
  int  *msg_id;				/* message id of irecv */ 
  int  i,j,k,m,n,p;
  int  count;
  int  *l;
  char sync_buf[1];			/* buffer for synchronizing messages */
  char *buffer, *ptr;
  int  *ibuf, *idata;
  double *dbuf, *ddata;
   
#ifdef DEBUG
  trace_enter("new_execute_gather");
#endif

  /* message id's for asynchronous communication */
  msg_id = wbuffer->msg_id;

  /* message buffer */
  if (wbuffer->max_size < size*narray*pattern->buffer_size) {
    buffer = (char *) malloc(size*narray*pattern->buffer_size);
    allocated = TRUE;
  }
  else {
    buffer = (char *) wbuffer->buffer;
  }

  /* check if processor send messages to itself */
  if (pattern->send_n[PARTI_mynode]) {
    /* copy data to buffer */
    l = pattern->send_list[PARTI_mynode];
    switch (size) {
    case sizeof(int):
      ibuf = (int *) buf_list[PARTI_mynode];
      for (i = 0, n = 0; i < narray; i++) {
        idata = (int *) data[i];
        for (j=0; j<pattern->send_n[PARTI_mynode]; j++, n++)
          ibuf[n] = idata[l[j]];
      }
      break;
    case sizeof(double):
      dbuf = (double *) buf_list[PARTI_mynode];
      for (i = 0, n = 0; i < narray; i++) {
        ddata = (double *) data[i];
        for (j=0; j<pattern->send_n[PARTI_mynode]; j++, n++)
          dbuf[n] = ddata[l[j]];
      }
      break;
    case sizeof(char):
      ptr = buf_list[PARTI_mynode];
      for (i = 0, n = 0; i < narray; i++) {
        for (j=0; j<pattern->send_n[PARTI_mynode]; j++, n++)
          ptr[n] = data[i][l[j]];
      }
      break;
    default:
      ptr = buf_list[PARTI_mynode];
      for (i = 0; i < narray; i++) {
        for (j=0; j<pattern->send_n[PARTI_mynode]; j++, ptr+=size)
	bytecopy(ptr, &data[i][l[j]*size], size);
      }
      break;
    }
  }
    
  /* call asynchronous receive funcations */
  for (p=0, count=0; p<PARTI_numnodes; p++){
    if ((pattern->fetch_n[p] > 0) && (p != PARTI_mynode)) {
      msg_id[count++] =
      PARTI_irecv (FORCED_GATHER_EXECUTOR_MSG+call_count*MAXNODES+p,
	           buf_list[p],pattern->fetch_n[p]*size*narray, p, PARTI_pid);
      PARTI_csend (GATHER_SYNC+call_count*MAXNODES+PARTI_mynode, 
		   sync_buf, 0, p, PARTI_pid);
    }
  }

  /* send data to other processors */
  for(i=1; i<PARTI_numnodes; i++){
    p = (PARTI_mynode + i) % PARTI_numnodes;
    if(pattern->send_n[p]){
      l = pattern->send_list[p];
      switch (size) {
      case sizeof(int):
        ibuf = (int *) buffer;
        for (j = 0, n = 0; j < narray; j++) {
          idata = (int *) data[j];
          for (k=0; k<pattern->send_n[p]; k++, n++)
            ibuf[n] = idata[l[k]];
        }
	break;
      case sizeof(double):
        dbuf = (double *) buffer;
        for (j = 0, n = 0; j < narray; j++) {
          ddata = (double *) data[j];
          for (k=0; k<pattern->send_n[p]; k++, n++)
            dbuf[n] = ddata[l[k]];
        }
	break;
      case sizeof(char):
        for (j = 0, n = 0; j < narray; j++) {
          for (k=0; k<pattern->send_n[p]; k++, n++)
            buffer[n] = data[j][l[k]];
        }
	break;
      default:
        for (j = 0, ptr=buffer; j < narray; j++) {
          for(k=0; k<pattern->send_n[p]; k++, ptr+=size)
            bytecopy(ptr,&data[j][l[k]*size],size);
        } 
	break;
      }
      PARTI_crecv (GATHER_SYNC+call_count*MAXNODES+p, sync_buf, 0, p, PARTI_pid);
      PARTI_bsend (FORCED_GATHER_EXECUTOR_MSG+call_count*MAXNODES+PARTI_mynode,
	     	   buffer, pattern->send_n[p]*size*narray,p,PARTI_pid);
    }  
  }
  call_count = (++call_count > MAXCOUNT)? 0: call_count;

  /* wait until all messages are done */
  for (i = 0; i < count; i++) {
    PARTI_msgwait (msg_id[i]);
  }

  /* free buffer is necessary */
  if (allocated)
    free (buffer);

#ifdef DEBUG
  trace_exit();
#endif
}

void 
new_multi_execute_scatter(pattern, func, size, narray, buf_list, data)
struct com_pattern *pattern;
void (*func)();
int  size;
int  narray;
char **buf_list,**data;
{
  static int call_count = 0;
  int  node;				/* processor id of incoming message */
  int  n_messages;			/* number of messages sent */
  int  allocated = FALSE;		/* new buffer allocated */
  int  *msg_id;				/* message id of isend */
  int  i,j,p;
  int  count;
  int  *l;
  char *ptr;
  char *buffer;
  char **wlist;				/* pointer of buffer */
  char sync_buf[1];			/* buffer for synchronizing messages */
  double end, start;

#ifdef DEBUG
  trace_enter("new_execute_scatter");
#endif


  /* message id's for asynchronous communication */
  msg_id = wbuffer->msg_id;

  /* message buffer */
  if (wbuffer->max_size < size*narray*pattern->total_send_size) {
    buffer = (char *) malloc(size*narray*pattern->total_send_size);
    allocated = TRUE;
  }
  else {
    buffer = (char *) wbuffer->buffer;
  }

  /* check if processor send messages to itself */
  if (pattern->fetch_n[PARTI_mynode]) {
    /* copy the buffer to target */
    for (i = 0, ptr = buf_list[PARTI_mynode]; i < narray; i++) {
      scatter_func(data[i], ptr, pattern->send_list[PARTI_mynode],
		   pattern->send_n[PARTI_mynode], size, func);
      ptr += pattern->send_n[PARTI_mynode] * size;
    }
  }

  /* set up the buffer list */
  wlist = wbuffer->wlist;
  for (p = 0, count = 0; p < PARTI_numnodes; p++) {
    wlist[p] = &buffer[count];
    if (p != PARTI_mynode)
      count += (pattern->send_n[p]*size*narray);
  }
    
  /* call asynchronous receive funcations */
  for(i=1, n_messages=0; i<PARTI_numnodes; i++){
    p = pattern->gcolor[i];
    if (pattern->send_n[p]) {
      msg_id[n_messages++] =
      PARTI_irecv (FORCED_SCATTER_EXECUTOR_MSG+call_count*MAXNODES+p,
	    	   wlist[p], pattern->send_n[p]*size*narray, p, PARTI_pid);
      PARTI_csend (SCATTER_SYNC+call_count*MAXNODES+PARTI_mynode, 
		   sync_buf, 0, p, PARTI_pid);
    }
    
    /* send data */
    if (pattern->fetch_n[p]) {
      PARTI_crecv (SCATTER_SYNC+call_count*MAXNODES+p, sync_buf, 0, p, PARTI_pid);
      PARTI_bsend (FORCED_SCATTER_EXECUTOR_MSG+call_count*MAXNODES+PARTI_mynode,
                   buf_list[p], size*pattern->fetch_n[p]*narray, p, PARTI_pid);
    }
  }
  call_count = (++call_count > MAXCOUNT)? 0: call_count;

  /* wait messages and store data */
  for (i = 1, count = 0; i < PARTI_numnodes; i++) {
    p = pattern->gcolor[i];
    if (pattern->send_n[p]) {
      PARTI_msgwait(msg_id[count]);
      for (j = 0, ptr = wlist[p]; j < narray; j++) {
        scatter_func(data[j], ptr, pattern->send_list[p], 
		     pattern->send_n[p], size, func);
        ptr += pattern->send_n[p] * size;
      }
      count++;
    }
  }

  /* free the buffer is necessary */
  if (allocated)
    free (buffer);

#ifdef DEBUG 
  trace_exit(); 
#endif    
}
